/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL
v2. You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "rm_file_handle.h"
#include "errors.h"
#include "storage/page.h"
/**
 * @description: 获取当前表中记录号为rid的记录
 * @param {Rid&} rid 记录号，指定记录的位置
 * @param {Context*} context
 * @return {unique_ptr<RmRecord>} rid对应的记录对象指针
 */
std::unique_ptr<RmRecord> RmFileHandle::get_record(const Rid &rid,
                                                   Context *context) const {
  // done:
  // 1. 获取指定记录所在的page handle
  RmPageHandle page_handle = fetch_page_handle(rid.page_no);
  // 根据rid.page_no和rid.slot_no计算出记录在页面中的偏移量
  char *record = page_handle.get_slot(rid.slot_no);
  // 2. 初始化一个指向RmRecord的指针（赋值其内部的data和size）
  std::unique_ptr<RmRecord> rm_record =
      std::make_unique<RmRecord>(page_handle.file_hdr->record_size, record);
  return rm_record;
}

/**
 * @description: 在当前表中插入一条记录，不指定插入位置
 * @param {char*} buf 要插入的记录的数据
 * @param {Context*} context
 * @return {Rid} 插入的记录的记录号（位置）
 */
Rid RmFileHandle::insert_record(char *buf, Context *context) {
  // by noah
  // 1. 获取当前未满的page handle(在没有空闲页的时候需要创建空闲页,所以page_handle一定是空闲的)
  RmPageHandle page_handle = create_page_handle();
  // 2. 根据bitmap在page handle中找到空闲slot位置
  int slot_no = Bitmap::first_bit(false, page_handle.bitmap,
                                 this->file_hdr_.num_records_per_page);
  // 3. 将buf复制到空闲slot位置
  char *record = page_handle.get_slot(slot_no);
  memcpy(record, buf, this->file_hdr_.record_size);
  // 4. 更新page_handle.page_hdr中的数据结构
  page_handle.page_hdr->num_records++;
  // 注意考虑插入一条记录后页面已满的情况，需要更新file_hdr_.first_free_page_no
  if (page_handle.page_hdr->num_records ==
      this->file_hdr_.num_records_per_page) {
    this->file_hdr_.first_free_page_no =
        page_handle.page_hdr->next_free_page_no;
  }
  return Rid{page_handle.page->get_page_id().page_no, slot_no};
}

/**
 * @description: 在当前表中的指定位置插入一条记录
 * @param {Rid&} rid 要插入记录的位置
 * @param {char*} buf 要插入记录的数据
 */
void RmFileHandle::insert_record(const Rid &rid, char *buf) {
  // 1. 获取指定记录所在的page handle
  RmPageHandle page_handle = fetch_page_handle(rid.page_no);
  // 判断当前位置上是否有记录
  if (!is_record(rid)) {
    // 根据rid.page_no和rid.slot_no计算出记录在页面中的偏移量
    char *record = page_handle.get_slot(rid.slot_no);
    // 3. 将buf复制到空闲slot位置
    memcpy(record, buf, this->file_hdr_.record_size);
    // 4. 更新page_handle.page_hdr中的数据结构
    page_handle.page_hdr->num_records++;
    // 注意考虑插入一条记录后页面已满的情况，需要更新file_hdr_.first_free_page_no
    if (page_handle.page_hdr->num_records ==
        this->file_hdr_.num_records_per_page) {
      this->file_hdr_.first_free_page_no =
          page_handle.page_hdr->next_free_page_no;
    }
  } else {
    throw RecordExistError();
  }
}

/**
 * @description: 删除记录文件中记录号为rid的记录
 * @param {Rid&} rid 要删除的记录的记录号（位置）
 * @param {Context*} context
 */
void RmFileHandle::delete_record(const Rid &rid, Context *context) {
  // done:
  if(is_record(rid)){
    // 1. 获取指定记录所在的page handle
    RmPageHandle page_handle = fetch_page_handle(rid.page_no);
    // 2. 更新page_handle.page_hdr中的数据结构
    page_handle.page_hdr->num_records--;
    // 更新位图
    Bitmap::reset(page_handle.bitmap, rid.slot_no);
    // 注意考虑删除一条记录后页面未满的情况，需要调用release_page_handle()
    if (page_handle.page_hdr->num_records ==
        this->file_hdr_.num_records_per_page - 1) {
        release_page_handle(page_handle);
    }
  }else{
    throw RecordNotExistError();
  }
}

/**
 * @description: 更新记录文件中记录号为rid的记录
 * @param {Rid&} rid 要更新的记录的记录号（位置）
 * @param {char*} buf 新记录的数据
 * @param {Context*} context
 */
void RmFileHandle::update_record(const Rid &rid, char *buf, Context *context) {
  // done:
  // 1. 获取指定记录所在的page handle
  RmPageHandle page_handle = fetch_page_handle(rid.page_no);
  // 更新记录一定要求原记录存在
  if (is_record(rid)) {
    // 2. 更新记录
    char *record = page_handle.get_slot(rid.slot_no);
    memcpy(record, buf, this->file_hdr_.record_size);
  } else {
    throw RecordNotExistError();
  }
}

/**
 * 以下函数为辅助函数，仅提供参考，可以选择完成如下函数，也可以删除如下函数，在单元测试中不涉及如下函数接口的直接调用
 */
/**
 * @description: 获取指定页面的页面句柄
 * @param {int} page_no 页面号
 * @return {RmPageHandle} 指定页面的句柄
 */
RmPageHandle RmFileHandle::fetch_page_handle(int page_no) const {
  // done:
  // 使用缓冲池获取指定页面，并生成page_handle返回给上层
  // if page_no is invalid, throw PageNotExistError exception
  Page *myPage;
  // 判断page_no的合法性
  if (this->file_hdr_.num_pages <= page_no || page_no < 0) {
    // 根据文件描述符获取记录文件的文件名
    std::string filename = this->disk_manager_->get_file_name(this->fd_);
    // 抛出PageNotExistError异常
    throw PageNotExistError(filename, page_no);
  } else {
    // 创建一个PageId
    PageId myPageId;
    myPageId.page_no = page_no;
    myPageId.fd = this->fd_;
    // 获取RmPageHandle
    myPage = this->buffer_pool_manager_->fetch_page(myPageId);
  }
  return RmPageHandle(&(this->file_hdr_), myPage);
}

/**
 * @description: 创建一个新的page handle
 * @return {RmPageHandle} 新的PageHandle
 */
RmPageHandle RmFileHandle::create_new_page_handle() {
  // by liamY
  // 1.使用缓冲池来创建一个新page
  // 2.更新page handle中的相关信息
  // 3.更新file_hdr_

  // 使用缓冲池创建一个新page
  PageId page_id;
  Page *new_page = this->buffer_pool_manager_->new_page(&page_id);
  if (new_page == nullptr) {
    throw UnixError();
  }
  // 更新page handle中的相关信息
  RmPageHandle page_handle = RmPageHandle(&file_hdr_, new_page);
   page_handle.page_hdr->next_free_page_no = -1; 
  // 更新file_hdr_
  file_hdr_.first_free_page_no = page_id.page_no;
  file_hdr_.num_pages++;
  return page_handle;
}

/**
 * @brief 创建或获取一个空闲的page handle
 *
 * @return RmPageHandle 返回生成的空闲page handle
 * @note pin the page, remember to unpin it outside!
 */
RmPageHandle RmFileHandle::create_page_handle() {
    // Todo:
    // 1. 判断file_hdr_中是否还有空闲页
    //     1.1
    //     没有空闲页：使用缓冲池来创建一个新page；可直接调用create_new_page_handle()
    //     1.2 有空闲页：直接获取第一个空闲页
    // 2. 生成page handle并返回给上层
    // // 获取第一个空闲页
    RmPageHandle page_handle1 = fetch_page_handle(this->file_hdr_.first_free_page_no);
    // 初始化时无page可用；在运行过程中可能最后一页恰好满了，此时first_free_page_no将会被更新
    if (this->file_hdr_.first_free_page_no == -1 || 
    Bitmap::first_bit(false, page_handle1.bitmap,this->file_hdr_.num_records_per_page)==this->file_hdr_.num_records_per_page) {
        RmPageHandle page_handle2 = create_new_page_handle();
        if (page_handle2.page != nullptr) {
            // 更新first_free_page_no
            this->file_hdr_.first_free_page_no =page_handle2.page->get_page_id().page_no;
            // 更新numpage
            this->file_hdr_.num_pages++;
        }
        return page_handle2;
    } else {
        // 返回第一个空闲页
        return page_handle1;
    }
}

/**
 * @description:
 * 当一个页面从没有空闲空间的状态变为有空闲空间状态时，更新文件头和页头中空闲页面相关的元数据
 */
void RmFileHandle::release_page_handle(RmPageHandle &page_handle) {
  // Todo:
  // 当page从已满变成未满，考虑如何更新：
  // 1. 更新page_handle.page_hdr->next_free_page_no
  page_handle.page_hdr->next_free_page_no = this->file_hdr_.first_free_page_no;
  // 2. 更新file_hdr_.first_free_page_no
  this->file_hdr_.first_free_page_no = page_handle.page->get_page_id().page_no;
}